// Sample code: Enroll templates into a collection, then run 1N identification

// This sample app demonstrates how to use 1N identification. First, a new collection is created.
// Next, templates are generated for several identities and enrolled into the collection.
// Finally, 1N identification is run to determine the identity of an anonymous template.
// To make the data enrolled in the collection persist, change the dbms configuration option
// Build on Linux: g++ -o identification --std=c++11 ./identification.cpp -I ../ -L ../trueface_sdk -l tf -fopenmp
// Build on macOS: g++ -o identification --std=c++11 ./identification.cpp -I ../ -L ../trueface_sdk -l tf

#include "trueface_sdk/tf_sdk.h"
#include <iostream>
#include <vector>
#include <string>

int main() {
    // For a full list of configuration options, visit: https://reference.trueface.ai/cpp/dev/latest/usage/general.html#_CPPv4N8Trueface20ConfigurationOptionsE
    Trueface::ConfigurationOptions options;
    options.frModel = Trueface::FacialRecognitionModel::LITE; // TODO: Can use full model here for better performance
    options.dbms = Trueface::DatabaseManagementSystem::NONE; // The data will not persist after the app terminates using this backend option.
    // TODO: If you have a NVIDIA gpu, then enable the enableGPU flag (you will require a GPU specific token for this).

    // If you want the data to persist, use SQLITE or POSTGRESQL option.
//    options.dbms = Trueface::DatabaseManagementSystem::SQLITE;
//    options.dbms = Trueface::DatabaseManagementSystem::POSTGRESQL;

    Trueface::SDK tfSdk(options);
    // TODO: replace the string with your license code.
    bool valid = tfSdk.setLicense("${LICENSE_CODE}");

    if (valid == false) {
        std::cout << "Error: the provided license is invalid." << std::endl;
        return 1;
    }

    // Create a new collection or load existing collection
    const std::string databaseName = "my_database.db";
    const std::string collectionName = "my_collection";

    // If using Trueface::DatabaseManagementSystem::NONE this does not need to be called
    // but doing so is a harmless no-op
    auto retcode = tfSdk.createDatabaseConnection(databaseName);

    // If using the POSTGRESQL backend, then it might look something like this...
//    auto retcode = tfSdk.createDatabaseConnection("host=localhost port=5432 dbname=my_database user=postgres password=admin");

    if (retcode != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Unable to create database connection\n";
        return -1;
    }

    // Create a new collection, or if a collection with the given name already exists in the database, load the collection into memory.
    retcode = tfSdk.createLoadCollection(collectionName);
    if (retcode != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Unable to create or load collection\n";
        return -1;
    }

    // Since our collection is empty, lets populate the collection with some identities
    std::vector<std::pair<std::string, std::string>> identitiesVec = {
            {"../images/armstrong_2.jpg", "Armstrong"},
            {"../images/armstrong_3.jpg", "Armstrong"}, // Can add the same identity more than once
            {"../images/aldrin.jpg", "Aldrin"}
    };

    for (const auto& identity: identitiesVec) {
        std::cout << "Processing image: " << identity.first << " with identity: " << identity.second << std::endl;
        // Generate a template for each image
        retcode = tfSdk.setImage(identity.first);
        if (retcode != Trueface::ErrorCode::NO_ERROR) {
            std::cout << "Unable to set the image\n";
            return -1;
        }

        // Detect the largest face in the image
        Trueface::FaceBoxAndLandmarks faceBoxAndLandmarks;
        bool faceDetected;
        auto errorCode = tfSdk.detectLargestFace(faceBoxAndLandmarks, faceDetected);
        if (errorCode != Trueface::ErrorCode::NO_ERROR) {
            std::cout << "There was an error with the call to detectLargestFace\n";
            continue;
        }

        if (!faceDetected) {
            std::cout << "Unable to detect face\n";
            continue;
        }

        // We want to only enroll high quality images into the database / collection
        // Therefore, ensure that the face height is at least 100px
        auto faceHeight = faceBoxAndLandmarks.bottomRight.y - faceBoxAndLandmarks.topLeft.y;
        std::cout << "Face height: " << faceHeight << std::endl;
        if (faceHeight < 100) {
            std::cout << "The face is too small in the image for a high quality enrollment." << std::endl;
            continue;
        }

        // Get the aligned face chip so that we can compute the image quality
        uint8_t alignedImage[37632];
        errorCode = tfSdk.extractAlignedFace(faceBoxAndLandmarks, alignedImage);
        if (errorCode != Trueface::ErrorCode::NO_ERROR) {
            std::cout << "There was an error extracting the aligned face\n";
            continue;
        }

        float quality;
        errorCode = tfSdk.estimateFaceImageQuality(alignedImage, quality);
        if (errorCode != Trueface::ErrorCode::NO_ERROR) {
            std::cout << "Unable to compute image quality\n";
            continue;
        }

        // Ensure the image quality is above a threshold (TODO: adjust this threshold based on your use case).
        // Once again, we only want to enroll only high quality images into our collection
        std::cout << "Face quality: " << quality << std::endl;
        if (quality < 0.8) {
            std::cout << "Please choose a higher quality enrollment image\n";
            return -1;
        }

        // As a final check, we can check the orientation of the head and ensure that it is facing forward
        // To see the effect of yaw and pitch on the match score, refer to: https://reference.trueface.ai/cpp/dev/latest/usage/face.html#_CPPv4N8Trueface3SDK23estimateHeadOrientationERK19FaceBoxAndLandmarksRfRfRf
        float yaw, pitch, roll;
        errorCode = tfSdk.estimateHeadOrientation(faceBoxAndLandmarks, yaw, pitch, roll);
        if (errorCode != Trueface::ErrorCode::NO_ERROR) {
            std::cout << "Unable to compute head orientation\n";
            return -1;
        }

        std::cout << "Yaw: " << yaw * 180 / 3.14  << ", Pitch: " << pitch * 180 / 3.14 << ", Roll: " << roll * 180 / 3.14 << " degrees" << std::endl;

        // TODO: Can filter out images with extreme yaw and pitch here

        // Generate the enrollment template
        Trueface::Faceprint enrollmentFaceprint;
        errorCode = tfSdk.getFaceFeatureVector(alignedImage, enrollmentFaceprint);
        if (errorCode != Trueface::ErrorCode::NO_ERROR) {
            std::cout << "Error: Unable to generate template\n";
            continue;
        }

        // Enroll the template into the collection
        std::string UUID;
        retcode = tfSdk.enrollTemplate(enrollmentFaceprint, identity.second, UUID);
        if (retcode != Trueface::ErrorCode::NO_ERROR) {
            std::cout << "Unable to enroll template\n";
            continue;
        }

        std::cout << "Success: Enrolled template with UUID: " << UUID << std::endl;
        // TODO: Can choose to store UUID for each faceprint which is enrolled into the collection
        // The UUID can later be used to delete the template from the collection

        std::cout << "----------------------------------" << std::endl;
    }

    // Use image of armstrong as probe
    retcode = tfSdk.setImage("../images/armstrong_1.jpg");
    if (retcode != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Unable to set the image\n";
        return -1;
    }

    // Generate a template as the probe
    // We do not need to ensure that the probe template is high quality.
    // Only the enrollment templates must be high quality.
    Trueface::Faceprint probeFaceprint;
    retcode = tfSdk.getLargestFaceFeatureVector(probeFaceprint);
    if (retcode != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Unable to generate template\n";
        return -1;
    }

    // Run identify function
    Trueface::Candidate candidate;
    bool found;

    // TODO: Select a threshold for your application using the ROC curves
    // https://performance.trueface.ai/
    const float threshold = 0.3;
    retcode = tfSdk.identifyTopCandidate(probeFaceprint, candidate, found, threshold);

    if (retcode != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "There was an error with the call to identify" << std::endl;
        return -1;
    }

    if (!found) {
        std::cout << "Unable to find match" << std::endl;
        return -1;
    }

    std::cout << "Identity found: " << candidate.identity << std::endl;
    std::cout << "Match Similarity: " << candidate.similarityMeasure << std::endl;
    std::cout << "Match Probability: " << candidate.matchProbability << std::endl;

    // For the sake of the demonstration, remove all the templates from the collection
    for (const auto& identity: identitiesVec) {
        retcode = tfSdk.removeByIdentity(identity.second);

        if (retcode != Trueface::ErrorCode::NO_ERROR) {
            std::cout << "Unable to remove from collection\n";
            return -1;
        }
    }

    return 0;
}


